# coding: utf-8
from __future__ import absolute_import
import six

from collections import defaultdict
from functools import reduce
from itertools import groupby
from operator import or_

from django.db import router
from django.db.models.deletion import Collector
from django.db.models.fields.related import RelatedField
from django.db.models.query_utils import Q
from django.dispatch import Signal
from m3.actions import OperationResult
from objectpack.actions import ObjectDeleteAction
from objectpack.models import ModelProxy

from educommon.utils.db import get_related_fields
from educommon.utils.db import get_related_instances

from .mixins import CascadeDeleteMixin
from .ui import CancelConfirmWindow


class DeleteCheck(object):

    u"""Проверяет объекты перед удалением на наличие зависимостей.

    В случае, если в системе есть объекты, зависимые от удаляемого объекта и
    не отмеченные для каскадного удаления через cascade_delete_objects, то
    отменяет удаление и выводит окно с предупреждением и списком зависимых
    объектов.
    Если в системе есть только зависимые объекты, помеченные
    cascade_delete_objects, то выводится сообщение с кнопкой Удалить, для
    удаления и связанных объектов.

    Для добавления объектов, которые *неявно* зависят от удаляемых, необходимо
    подписаться на сигнал ``DeleteCheck.collect_implicit``.
    """

    #: Сигнал отправляемый после сборки зависимых объектов
    #:
    #: :param objects_to_delete: Удаляемые объекты.
    #: :type objects_to_delete: django.db.models.query.QuerySet
    #: :param related_objects: Словарь, сопоставляющий удаляемые
    #:     и зависимые объекты.
    #: :type related_objects: collections.defaultdict
    collect_implicit = Signal(
        providing_args=['objects_to_delete', 'related_objects']
    )

    def _get_message(self, objects):
        """Возвращает сообщение в правильном склонении."""
        return (
            u'объекта' if len(objects) == 1 else u'объектов',
            u'него' if len(objects) == 1 else u'них'
        )

    def _setup_window(self, objects, related_objects, blocked=True):
        """Возвращает окно с сообщением о связанных объектах.

        :param objects: удаляемые объекты
        :param related_objects: связанные объекты
        :param bool blocked: режим показа окна блокирующее/не блокирующее
        :return: CancelConfirmWindow
        """
        if blocked:
            title = (
                u'Удаление {} невозможно, т.к. на {} есть ссылки:'
                .format(*self._get_message(objects))
            )
        else:
            title = (
                u'При удалении {} будут удалены следующие связи:'
                .format(self._get_message(objects)[0])
            )

        win = CancelConfirmWindow()
        win.set_params(dict(
            title=title,
            objects=objects,
            related_objects=related_objects,
            blocked=blocked
        ))
        win.pack_action_url = self.action.get_absolute_url()
        return win

    def _get_objects(self, context):
        model = self.action.parent.model
        if issubclass(model, ModelProxy):
            model = model.model

        object_ids = getattr(context, self.action.parent.id_param_name)

        return model.objects.filter(pk__in=object_ids)

    def _get_related_instances(self, obj, skip_func=None):
        if isinstance(obj, ModelProxy):
            obj = getattr(obj, obj.model.__name__.lower())
        return get_related_instances(obj, collapse=False, skip_func=skip_func)        

    def _get_non_block_relations(self, obj):
        """Возвращает неблокирующие удаление связи."""
        displayed_relations = dict()
        collector = Collector(
            using=router.db_for_write(obj.__class__, instance=obj)
        )
        collector.collect((obj,))
        # Коллектор содержит ссылку на obj, удалим сам объект из коллектора 
        collector.data[obj.__class__].remove(obj)
        m2m_through_models = set(
            m2m.remote_field.through 
            for m2m in obj.__class__._meta.local_many_to_many
        )

        for model, related_objects in six.iteritems(collector.data):
            if all((
                getattr(model, 'display_related_error', True),
                model not in m2m_through_models,
                related_objects
            )):
                key = '.'.join((
                    model._meta.app_label,
                    model._meta.object_name,
                ))
                displayed_relations[key] = related_objects
        return displayed_relations

    def before(self, request, context):
        if not isinstance(self.action, ObjectDeleteAction):
            return

        # Удаляемые объекты
        objects = self._get_objects(context)

        if not objects:
            return

        if hasattr(self.action.parent, 'skip_field'):
            def skip_function(field):
                return (
                    self.action.parent.skip_field(field) or
                    CascadeDeleteMixin.skip_field(field)
                )
        else:
            skip_function = CascadeDeleteMixin.skip_field

        if getattr(context, 'delete_check_confirmed', False):
            for obj in objects:
                obj.delete()
            return

        # Объекты, зависящие от удаляемых
        related_objects = defaultdict(list)

        # Проверка блокирующих удаление связей
        for obj in objects:
            for related_obj in self._get_related_instances(obj, skip_function):
                key = '.'.join((
                    related_obj._meta.app_label,
                    related_obj._meta.object_name,
                ))
                related_objects[key].append(related_obj)
        self.collect_implicit.send(
            self.action.parent.model,
            objects_to_delete=objects,
            related_objects=related_objects
        )

        if related_objects:
            win = self._setup_window(objects, related_objects)
        else:
            # Проверка не блокирующих удаление связей
            for obj in objects:
                related_objects.update(self._get_non_block_relations(obj))
            win = self._setup_window(objects, related_objects, blocked=False)

        if not related_objects:
            return 

        win.action_context = context
        return OperationResult(
            success=False,
            code=win.get_script(),
        )
